>{-# OPTIONS -fglasgow-exts #-}

> module StateArrow(StateArrow) where

> import Arrow
> import ArrowTransformer
> import ArrowOps

> newtype StateArrow s a b c = ST (a (b, s) (c, s))
> unST (ST f) = f

> swapsnd ~(~(x, y), z) = ((x, z), y)

> instance Arrow a => Arrow (StateArrow s a) where
>	arr f = ST (arr (\(x, s) -> (f x, s)))
>	ST f >>> ST g = ST (f >>> g)
>	first (ST f) = ST (arr swapsnd >>> first f >>> arr swapsnd)

> instance Arrow a => ArrowTransformer (StateArrow s) a where
>	lift f = ST (first f)

> instance ArrowFunctor (StateArrow s) where
>	arrowmap embed (ST f) = ST (embed f)

operations

> instance Arrow a => ArrowState s (StateArrow s a) where
>	fetch = ST (arr (\(_, s) -> (s, s)))
>	store = ST (arr (\(s, _) -> ((), s)))

> instance Arrow a => ArrowRunState s (StateArrow s a) a where
>	runState (ST f) = f

The following promotions follow directly from the arrow transformer.

> instance (ArrowZero a) => ArrowZero (StateArrow s a) where
>	zeroArrow = lift zeroArrow

> instance (ArrowCircuit a) => ArrowCircuit (StateArrow s a) where
>	delay x = lift (delay x)

> instance (ArrowExcept ex a) => ArrowExcept ex (StateArrow s a) where
>       raise = lift raise

> instance (ArrowReader r a) => ArrowReader r (StateArrow s a) where
>       readState = lift readState

> instance (ArrowWriter w a) => ArrowWriter w (StateArrow s a) where
>       write = lift write

liftings of standard classes

> instance ArrowChoice a => ArrowChoice (StateArrow s a) where
>	left (ST f) = ST (arr distr >>> left f >>> arr undistr)
>		where	distr (Left y, s) = Left (y, s)
>			distr (Right z, s) = Right (z, s)
>			undistr (Left (y, s)) = (Left y, s)
>			undistr (Right (z, s)) = (Right z, s)

> instance ArrowApply a => ArrowApply (StateArrow s a) where
>	app = ST (arr (\((ST f, x), s) -> (f, (x, s))) >>> app)

> instance ArrowLoop a => ArrowLoop (StateArrow s a) where
>	loop (ST f) = ST (loop (arr swapsnd >>> f >>> arr swapsnd))

> instance ArrowPlus a => ArrowPlus (StateArrow s a) where
>	ST f <+> ST g = ST (f <+> g)

promotions

> instance ArrowRunReader r a a' =>
>		ArrowRunReader r (StateArrow s a) (StateArrow s a') where
>	runReader (ST f) = ST (arr swapsnd >>> runReader f)

> instance ArrowNewReader r a => ArrowNewReader r (StateArrow s a) where
>	newReader (ST f) = ST (arr swapsnd >>> newReader f)

> instance ArrowRunWriter w a a' =>
>		ArrowRunWriter w (StateArrow s a) (StateArrow s a') where
>	runWriter (ST f) = ST (runWriter f >>> arr swapsnd)

> instance ArrowNewWriter w a => ArrowNewWriter w (StateArrow s a) where
>	newWriter (ST f) = ST (newWriter f >>> arr swapsnd)

> instance ArrowRunExcept ex a a' =>
>		ArrowRunExcept ex (StateArrow s a) (StateArrow s a') where
>	runExcept (ST f) (ST h) = ST (runExcept f (arr swapsnd >>> h))

> instance ArrowNewExcept ex a => ArrowNewExcept ex (StateArrow s a) where
>	newExcept (ST f) (ST h) = ST (newExcept f (arr swapsnd >>> h))

> instance ArrowRunMaybe a a' =>
>		ArrowRunMaybe (StateArrow s a) (StateArrow s a') where
>	runMaybe (ST f) (ST h) = ST (runMaybe f h)
