module SubExpressions(SubExp,Location,
                      subExps, replace, 
                      abstractall, abstractnone,cse,fullcse,
                      abstractallthree)
where

import List
import Utils
import Names
import Types
import Expressions
import Substitutions
import PPrint
import Constants
import Syntax
import Infer
import Monad (foldM)
-- import Trace
import Maybe

notrace _ = id

trace = notrace

type SubExp = (Closure,Location)
type Location = [Direction]
data Direction = First | Second | Body deriving (Eq, Ord, Show)

first (e,x) = (e,First:x)
second (e,x) = (e,Second:x)
body (e,x) = (e,Body:x)

subExps :: Closure -> [SubExp]
subExps c = (c,[]):subpaths c 
        where subpaths (Var _ _,_) = []
              subpaths (Con _ _,_) = []
              subpaths (Ap e1 e2 _,bs) = map second (subExps (e2,bs)) ++
                                         map first (subExps (e1,bs))
              subpaths (Lam p e _,bs) = map body (subExps(e,tfreevars p ++ bs))


type CSubExp = (Closure,[[Direction]])

cleanSubExps :: (Exp -> Bool) -> Exp -> [CSubExp]
cleanSubExps t e = [(s,ys) | (s,xs) <- collect $ map free
                                          $ filter isclosed 
                                          $ filter (typematch t) 
                                          $ subExps( e,[]),
                              ys <- subsplus (nub xs) ]

type SubSubExp = (Exp,[Location],[(Exp,[Location])])

showse :: (Exp,[Location]) -> String
showse (e,locs) = "(" ++ prettyExp e ++ "," ++ show locs ++ ")"

showses = joinWith ";" . map showse

showsse :: SubSubExp -> String

showsse (e,locs,ses) = showse (e,locs) ++ ":" ++ showses ses

showsses :: [SubSubExp] -> String
showsses = joinWith "\n" . map showsse

showexps = joinWith "," . map prettyExp

subsubexps :: Exp -> [SubSubExp]

subsubexps (Var x t) = [(Var x t,[[]],[(Var x t,[[]])])]
subsubexps (Con c t) = [(Con c t,[[]],[(Con c t,[[]])])]

subsubexps (Lam p b t) = 
              (Lam p b t,[[]],(Lam p b t,[[]]):subs):subsubs
           where subsubs = map ssbody (subsubexps b)
                 subs = (map sbody.filter (not.invalid.fst).
                                           (\(_,_,z)->z).head) subsubs
                 ssbody (se,locs,sses) = (se,map (Body:) locs,sses)
                 sbody (se,locs) = (se,map (Body:) locs)
                 invalid (Con _ _) = False
                 invalid (Var y _) = y `elem` freevars p
                 invalid (Lam _ b _) = invalid b
                 invalid (Ap e1 e2 _) = invalid e1 || invalid e2


-- subsubexps e@(Lam _ _ _) = error ("subsubexps called with invalid expression: "
--                                 ++ prettyExp e)

subsubexps (Ap e1 e2 t) = (Ap e1 e2 t,[[]],subs):subsubs
           where subsubs1 = map ssfirst (subsubexps e1)
                 subsubs2 = map sssecond (subsubexps e2)
                 subsubs = foldl (flip sscollapsedup) [] (subsubs1++subsubs2)
                 subs1 = (map sfirst.(\(_,_,z)->z).head) subsubs1
                 subs2 = (map ssecond.(\(_,_,z)->z).head) subsubs2
                 subs = (Ap e1 e2 t,[[]]):foldl (flip scollapsedup) [] 
                                                          (subs1++subs2)
                 ssfirst (se,locs,sses) = (se,map (First:) locs,sses)
                 sssecond (se,locs,sses) = (se,map (Second:) locs,sses)
                 sfirst (se,locs) = (se,map (First:) locs)
                 ssecond (se,locs) = (se,map (Second:) locs)
                 scollapsedup (se,locs) [] = [(se,locs)]
                 scollapsedup (se1,locs1) ((se2,locs2):ss)
                        = if se1==se2 then (se1,locs1++locs2):ss
                          else (se2,locs2):scollapsedup (se1,locs1) ss
                 sscollapsedup (se,locs,sses) [] = [(se,locs,sses)]
                 sscollapsedup (se1,locs1,sses1) ((se2,locs2,sses2):ss)
                        = if se1==se2 then (se1,locs1++locs2,sses1):ss
                          else (se2,locs2,sses2):
                                      sscollapsedup (se1,locs1,sses1) ss

showm :: Show a => (a,b) -> String
showm = show.fst
 
findoccurrences (hint,(pvars,tvars)) e = 
   trace ("fo: "++show pvars ++ ":" ++ show tvars ++ "\n") $
    [trace ("3sm"++showm match++"\n") $
       (zip (map (getvar match) bs) varlocs,locs)
                      | (se,locs,_) <- subsubexps e,
                         match <- threesimplematch (pat,pvars) (se,tvars)]
             where (pat,bs) = getbody hint
                   getvar subst (x,t) = appSubst subst (Var x t)
                   findvar x (Var y _) = if x==y then [[]] else []
                   findvar x (Con c _) = []
                   findvar x (Ap e1 e2 _) = map (First:) (findvar x e1) 
                                              ++ map (Second:) (findvar x e2)
                   findvar x (Lam _ b _) = map (Body:) (findvar x b)
                   varlocs = map (flip findvar pat.fst) bs

threesimplematch :: (Exp,[VarId]) -> (Exp,[VarId]) -> [Subst]

threesimplematch (p,pvars) (t,tvars) =
       threesm idSubst [(p,[],t,[])]
   where 
        threesm :: Subst -> [(Exp,[VarId],Exp,[VarId])] -> [Subst]
        threesm sub [] = [sub]
        threesm sub ((Var x s,xs,e,ys):es) 
          = 
            if x `elem` xs++pvars
            then (case e of
                   (Var y t) -> if (y `elem` ys++tvars 
                                 && posin (xs++pvars) x == posin (ys++tvars) y)
                                then threesmm sub (singletype s t) es
                                else []
                   _ -> [])
            else if null (freevars e `intersect` ys)
                 then threesmm sub (singleSubst x s e) es
                 else []

        threesm sub ((Con c s,_,Con d t,_) : es) 
          = if c==d
            then threesmm sub (singletype s t) es
            else []
        threesm sub ((Lam p d s,xs,Lam q e t,ys) : es) 
          = threesmm sub us es'
            where us = singletype s t
                  nxs = freevars p ++ xs
                  nys = freevars q ++ ys 
                  es' = (p,nxs,q,nys):(d,nxs,e,nys):es
        threesm sub ((Ap f1 a1 t1,xs,Ap f2 a2 t2,ys): es) 
          = threesmm sub (singletype t1 t2) ((f1,xs,f2,ys):(a1,xs,a2,ys):es)
        threesm _ _ = [] 
        threesmm sub us es = 
            concat [threesm (compose u sub) (map (papply u) es) | u <- us]



abstractallthree :: Exp -> (Exp,([VarId],[VarId])) -> Name [Exp]

abstractallthree e (hint,(pvars,tvars))
  = do x <- freshVar
       return $ nub $ map (betacreate x) $ map etaRed $
                 (abstractthreerec x (hint,(pvars,tvars)) e)
   where betacreate x e = (\ret -> trace ("aa3: "++prettyExp ret++"\n") ret)
                             $ mapback $ apl (lambda (x,exptype hint) e) hint
         mapback (Var x t) = if x `elem` pvars 
                              then Var (tvars!!(fromJust (posin pvars x))) t
                              else Var x t
         mapback (Con c t) = Con c t
         mapback (Lam p b t) = Lam p (mapback b) t
         mapback (Ap e1 e2 t) = Ap (mapback e1) (mapback e2) t
	 etaRed (Lam (Var x _) (Ap e (Var y _) _) _) 
		| x==y && not (x `elem` freevars e) = e
	 etaRed e = e
               

abstractthreerec x hintvm exp = 
    exp:concat (map (abstractthreerec x hintvm) 
                   (abstractthreeonce x hintvm exp))

abstractthreeonce x (hint,varmap) exp = [replace exp absloc se |
            (sels,locs) <- findoccurrences (hint,varmap) exp,
            absloc <- locs,
            se <- [foldl apl (Var x (exptype hint)) (map fst sels)]]

multilocs :: [Location] -> [[Location]]

multilocs = filter (not.null) . multilocse

multilocse [] = [[]]
multilocse (loc:locs) = ml ++ map (loc:) ml
                 where ml = multilocse locs

joinWith s [] = []
joinWith s xs = foldl1 (\x y -> x++s++y) xs


free ((e,bs),l) = ((e,[]),l)

isclosed :: SubExp -> Bool
isclosed ((e,bs),l) = null (intersect (tfreevars e) bs)

typematch :: (Exp -> Bool) -> SubExp -> Bool
typematch t ((e,bs),l) = t e 


replace :: Exp -> Location -> Exp -> Exp
replace e1 [] e = e
replace (Ap e1 e2 t) (First:ds) e = Ap (replace e1 ds e) e2 t
replace (Ap e1 e2 t) (Second:ds) e = Ap e1 (replace e2 ds e) t
replace (Lam p e1 t) (Body:ds) e = Lam p (replace e1 ds e) t
replace e l r = error ("invalid replace: " ++ prettyExp e ++ " " ++ show l)

select :: Exp -> Location -> Exp
select e1 [] = e1
select (Ap e1 e2 t) (First:ds) = select e1 ds
select (Ap e1 e2 t) (Second:ds)  = select e2 ds
select (Lam p e1 t) (Body:ds) = select e1 ds 
select e l = error "invalid select"


abstract :: Exp -> CSubExp -> Name Exp
abstract e ((f,bs),ls) 
   = do x <- freshVar 
        f' <- renaming scrub f
        let r = foldr (flip apl) (Var x t') (map varid bs)
        let step l e = replace e l r
        let b = foldr step e ls
        return (Ap (lambda (x,t') b) f' t)
            where t = exptype e   
                  t' = exptype f

getbody :: Exp -> Closure
getbody (Lam (Var x t) b _) = 
    let (body,bounds)=getbody b in (body,(x,t):bounds) 
getbody e = (e,[])


abstractall :: Exp -> (Exp -> Bool) -> Name [Exp]
abstractall e t = mapM (abstract e) (filter nontriv cse) 
                  where nontriv (_,[[Second]]) = False
                        nontriv _ = True
                        cse = cleanSubExps t e

abstractnone :: Exp -> Type -> Name Exp
abstractnone e1 t = do x <- freshVar
                       let s = exptype e1
                       return (Lam (Var x t) e1 (arrow t s))

-- common subexpression elimination

cse :: Closure -> Name Closure 
cse (e,xs) = do e' <- foldM elim (Let (DList NoDef) e (exptype e)) cs
                return (unlet (revdefs e'),xs)
             where cs = maximals $
                        map normalise $
                        topsort $
                        filter mult $
                        collect $
                        map free $
                        filter (notfunc . exptype . getexp) $
                        filter nontriv $
                        filter isclosed $
                        subExps (e,xs)
                   unlet (Let (DList NoDef) e t) = e
                   unlet e = e
                   free ((e,bs),l) = (e,l)
                   nontriv ((Var _ _,_),_) = False
                   nontriv ((Con _ _,_),_) = False
                   nontriv _ = True
                   notfunc (TCon "Arrow" _) = False
                   notfunc _ = True
                   getexp ((e,_),_) = e
                   mult (_,ys) = length ys > 1

revdefs (Let (DList ds) e t)  
    = Let (DList (rev ds NoDef)) e t
           where rev NoDef y = y
                 rev (OneDef d ds n) es = rev ds (OneDef d es n)

elim (Let (DList ds) e t) (s,ys) 
    = do x <- freshVar
         let v = Var x (exptype s)
         return (Let (DList (OneDef (Fun v (select e (head ys))) ds 0))
                                           (foldr (repl v) e ys) t) 
           where repl v l e = replace e l v 

topsort [] = []
topsort es = sink : topsort (es \\ [sink])
             where sink = head [ (e,xs) | (e,xs) <- es, 
                                          and (map (not.precedes xs . snd) 
                                                  (es \\ [(e,xs)])) ]
precedes xs ys = or [isPrefix x y | x <- xs, y <- ys]

isPrefix [] xs = True
isPrefix (x:xs) (y:ys) = x==y && isPrefix xs ys
isPrefix _ _ = False

normalise (e,xs) = (e,sort xs)

maximals [] = []
maximals ((e,xs):es) = 
    if not (or (map (subsumes xs . snd) es)) 
    then (e,xs):maximals es
    else maximals es

subsumes xs ys = length xs == length ys && and (zipWith isPrefix ys xs)

cses :: Closure -> Name Closure 
cses (Var x t,bs) = return (Var x t,bs)
cses (Con c t,bs) = return (Con c t,bs)
cses (Ap e1 e2 t,bs) = do (e1',_) <- cses (e1,bs)
                          (e2',_) <- cses (e2,bs)
                          return (Ap e1' e2' t,bs)
cses (Lam p e t,bs) = do e' <- cse (e,bs)
                         (e'',_) <- subcses e'
                         return (Lam p e'' t,bs)

fullcse c = do c' <- cse c
               subcses c'

subcses :: Closure -> Name Closure
subcses (Let (DList ds) e t,bs) = do (e',_) <- cses (e,bs)
                                     return (Let (DList ds) e' t,bs)
subcses (e,bs) = cses (e,bs)


