// PanFactory.cpp: implementation of the CPanFactory class.
//
//////////////////////////////////////////////////////////////////////

#include "stdafx.h"
#include "PanFactory.h"
#include "Box.h"
#include "PanSlider.h"
#include "PanISlider.h"
#include "PanViewWindow.h"
#include "PanRadioBox.h"
#include "PanButton.h"
#include "PanFileSelector.h"
#include "PanTick.h"
#include "PanTextBox.h"

// ToDo: eliminate the need for this conversion.
static inline bool BOOL2bool(BOOL b) { return b ? true : false; }

//////////////////////////////////////////////////////////////////////
// Construction/Destruction
//////////////////////////////////////////////////////////////////////

CPanFactory::CPanFactory(CComQIPtr<IPanImage> pImage,
                         int *pDisplayPasses,
                         double *pMaxJitter,
                         CComBSTR docURL)
 : m_imagePtr(pImage), m_pDisplayPasses(pDisplayPasses), m_pMaxJitter(pMaxJitter),
   m_docURL(docURL)
{

}

CPanFactory::~CPanFactory()
{

}

//////////////////////////////////////////////////////////////////////
// Methods
//////////////////////////////////////////////////////////////////////

CWindowVirt*
CPanFactory::CreateUI (HWND hWndParent, UI* pUI)
{
    USES_CONVERSION;

    switch (pUI->kind) {
    case HBox:
    {
        CBox* box;
        CWindowVirt* ch;
        int i;

        // Creating the box - the 'magic' numbers specifying the size
        // and position will be updated once all the children have been
        // added and the size of the box has been recalculated.
        box = new CBox(false);

        box->Create(A2T(pUI->info.hbox->label),50,50,hWndParent);

        for (i=0;i < pUI->info.hbox->noChildren; i++) {
            ch = CreateUI(*box, pUI->info.hbox->children[i]);
            box->AddChild(ch);
        }
        box->CalcLayout();
        return box;

    }
    case VBox:
    {
        CBox* box;
        CWindowVirt* ch;
        int i;

        // Creating the box - the 'magic' numbers specifying the size
        // and position will be updated once all the children have been
        // added and the size of the box has been recalculated.
        box = new CBox(true);

        box->Create(A2T(pUI->info.hbox->label),50,50,hWndParent);

        for (i=0;i < pUI->info.hbox->noChildren; i++) {
            ch = CreateUI(*box, pUI->info.hbox->children[i]);
            box->AddChild(ch);
        }
        box->CalcLayout();
        return box;

    }
    case Display:
    {
        CPanViewWindow* pWindow;

        pWindow = new CPanViewWindow (
            hWndParent, m_imagePtr, BOOL2bool(pUI->info.display->timeVarying),
            pUI->info.display->nWidth, pUI->info.display->nHeight,
            m_pDisplayPasses, m_pMaxJitter,
            m_perfData, m_mouseUpdater
            );
        m_viewInvalidator->SetView(pWindow);
        *m_pView = pWindow;
        return pWindow;

    }
    case Slider:
    {
        CPanSlider* slider;

        slider = new CPanSlider(hWndParent,
                                A2T(pUI->info.slider->label),
                                pUI->info.slider->min,
                                pUI->info.slider->max,
                                pUI->info.slider->pUpdateValue,
                                pUI->info.slider->pParam,
                                m_viewInvalidator);
        return slider;
    }
    case ISlider:
    {
        CPanISlider* slider;

        slider = new CPanISlider(hWndParent,
                                 A2T(pUI->info.islider->label),
                                 pUI->info.islider->min,
                                 pUI->info.islider->max,
                                 pUI->info.islider->pUpdateValue,
                                 pUI->info.islider->pParam,
                                 m_viewInvalidator);
        return slider;
    }
    case Radio:
    {
        CPanRadioBox* rBox;
        rBox = new CPanRadioBox(hWndParent,
                                A2T(pUI->info.radio->label),
                                pUI->info.radio->labels,
                                pUI->info.radio->noItems,
                                pUI->info.radio->defValue,
                                pUI->info.radio->pUpdateValue,
                                pUI->info.radio->pParam,
                                m_viewInvalidator);
        return rBox;
    }
    case Button:
    {
        CPanButton* btn;

        btn = new CPanButton(hWndParent,
                             A2T(pUI->info.button->label),
                             pUI->info.button->value,
                             50,30);

        return btn;
    }
    case Tick:
    {
        CPanTick* tick;

        tick = new CPanTick(hWndParent,
                            A2T(pUI->info.tick->label),
                            50,20,
                            pUI->info.tick->defValue,
                            pUI->info.tick->pUpdateValue,
                            pUI->info.tick->pParam,
                            m_viewInvalidator);
        return tick;
    }
    case ImportImage:
    {
        CPanImportImage* pImg;

        pImg = new CPanImportImage(hWndParent,
                                   A2T(pUI->info.image->label),
                                   A2T(pUI->info.image->defImg),
                                   70,20,
                                   pUI->info.image->pUpdateValue,
                                   pUI->info.image->pParamBits,
                                   pUI->info.image->pParamWidth,
                                   pUI->info.image->pParamHeight,
                                   m_viewInvalidator,
                                   m_docURL);
        m_stringProp->push_back(pImg);
        return pImg;
    }
    case TextBox:
    {
        CPanTextBox* pBox;

        pBox = new CPanTextBox(hWndParent,
                               A2T(pUI->info.textBoxInfo->label),
                               A2T(pUI->info.textBoxInfo->defVal),
                               70,20,
                               pUI->info.textBoxInfo->pUpdateValue,
                               pUI->info.textBoxInfo->pParamBits,
                               pUI->info.textBoxInfo->pParamWidth,
                               pUI->info.textBoxInfo->pParamHeight,
                               m_viewInvalidator);
        m_fontProp->push_back(pBox);
        m_stringProp->push_back(pBox);
        return pBox;
    }
    case Resizeable:
    {
        CWindowVirt* win = CreateUI (hWndParent, pUI->info.resize->uiElem);
        if (pUI->info.resize->hResize == FALSE) {
            SetProp(*win, RESIZE_PROPERTY_HORIZ, (HANDLE)RESIZE_NONE);
        }

        if (pUI->info.resize->vResize == FALSE) {
            SetProp(*win, RESIZE_PROPERTY_VERT, (HANDLE)RESIZE_NONE);
        }

        return win;
    }
    case MouseInput:
    {
        m_mouseUpdater->AddMouseListener(pUI->info.mouseInput->updMouse,
                                         pUI->info.mouseInput->pParamX,
                                         pUI->info.mouseInput->pParamY,
                                         m_viewInvalidator);
        return NULL;

    }
    default:
        // Always make sure the switch statement handles all permitted UI types
        // in uispec.h
        ::MessageBox(NULL, TEXT("UI element type not handled"), TEXT("CPanFactory::CreateUI"), MB_OK);
        return NULL;
    }
}

static
void
freeUI(UI* pUI);

CWindowVirt*
CPanFactory::Create(HWND hWnd,
                    std::vector<CStringProp*>* propVec,
                    std::vector<CFontProp*>* propFontVec,
                    CPerfData* pData,
                    CViewWindow** pView)
{
    UI* pUI;
    IPanImageUIPtr pUIPtr;
    CWindowVirt* winUI;
    CViewInvalidator* pInvalid = new CViewInvalidator();
    CMouseUpdater* pMouseUpdate = new CMouseUpdater();
    HRESULT hr;

    hr = m_imagePtr->QueryInterface(&pUIPtr);

    if (SUCCEEDED(hr)) {

        pUIPtr->GetUI((void**)&pUI);

        // Instead of passing the state needed when instantiating the UI tree as
        // arguments to CreateUI(), store it in object fields. The result is
        // a simpler calling interface & a solution that is more robust to change.
        // (You could argue that what has been done is the introduction of a 
        // state monad here, avoiding the plumbing of state.)

        m_mouseUpdater = pMouseUpdate;
        m_viewInvalidator = pInvalid;
        m_pView = pView;
        m_perfData = pData;
        m_stringProp = propVec;
        m_fontProp = propFontVec;

        winUI = CreateUI(hWnd, pUI);

        // Let go of the UI struct.
        pUIPtr->FreeUI(pUI);
        return winUI;
    } else {
        return NULL;
    }
}

