/*
 * Helper function for the (deprecated) use of 'importImage' in components.
 * 
 */
#include <windows.h>
#include <ocidl.h>
#include <olectl.h>

#define HIMETRIC_INCH	2540

extern "C" {

HRESULT LoadPictureFile(LPCTSTR szFile, IPicture** ppIPicture, int* pWidth, int* pHeight);

//
// FUNCTION:  ResolveImageName
//
// PURPOSE:   If the name of an image import doesn't refer to
//            an existing file, try prepending the directory
//            path of a the 'standard' image import directory.
//
//            If that doesn't refer to an existing file either, return
//            (a copy) of the original name. 
//
// COMMENTS:  The caller is responsible for free()ing the returned string.
//
char*
ResolveImageName(const char* szFile)
{
  char* fullName;
   
  if ( strchr(szFile, '\\') == NULL &&
       strchr(szFile, '/')  == NULL ) {
    // only do this for filenames that aren't absolute or relative
    // paths.
    HKEY  panKey;
    static char* imageDirStr = NULL;
    static int memoised = 0;
    DWORD dirStrLen;
    LONG  lRes;
	
    if (!memoised) {
      // Open our (machine-wide) location in the registry:
      lRes = RegOpenKeyEx
	  ( HKEY_LOCAL_MACHINE
	  , TEXT("Software\\Pan")
	  , 0
	  , KEY_READ
	  , &panKey
          );
      if ( lRes == ERROR_SUCCESS ) {

	// To avoid looking up the registry value everytime 
	// LoadImportedImg() is called, we memoise it here.
	//
	// The downside of this is, of course, that dynamic changes
	// to the imageDir value won't get picked up. There's ways
	// to get around that (cf. RegNotifyChangeValue()), but doesn't
	// seem to be worth the extra bother at the moment.
	//
	// We also leak a little bit of memory
	if (imageDirStr == NULL) {
	  imageDirStr = (char*)malloc(sizeof(char) * 256);
	}
	dirStrLen = 255;

	// Look for "imageDir" string value
	lRes = 
	  RegQueryValueEx ( panKey
			  , TEXT("imageDir")
			  , NULL
			  , NULL
			  , (BYTE*)imageDirStr
			  , &dirStrLen
			  );
	if ( lRes == ERROR_SUCCESS ) {
	  memoised = 1;
	}

	RegCloseKey(panKey);

      } else {
	// Make a copy of the string
	fullName = (char*)malloc(sizeof(char)*(strlen(szFile) + 1));
	strcpy(fullName, szFile);
    
	return fullName;
      }
    }

    // Prepend the imageDir to the filename.
    fullName = (char*)malloc(sizeof(char) * (strlen(imageDirStr) + strlen(szFile) + 2));

    strcpy(fullName, imageDirStr);
    strcat(fullName, "\\");
    strcat(fullName, szFile);
	
    return fullName;

  } else {
    // The filename contained forward or backward slashes, so 
    // assume it is pointing to some legitimate file.
    
    // Make a copy of the string
    fullName = (char*)malloc(sizeof(char)*(strlen(szFile) + 1));
    strcpy(fullName, szFile);
    
    return fullName;
  }
}

int
LoadImportedImg
         ( /*[in]*/   char* lpszFile
	 , /*[out]*/  int*  pWidth
	 , /*[out]*/  int*  pHeight
	 )
{
  IPicture *ppic;
  HBITMAP hBitmap;
  HRESULT hr;
  int width,height;
  HDC hdc = GetDC(NULL);
  short picType;
  RGBQUAD* ppv;
  char*    szFileReal;

  BITMAPINFO bmi;
  BITMAPINFOHEADER bmih;
  
  if (pWidth) {
    *pWidth = 0;
  }

  if (pHeight) {
    *pHeight = 0;
  }

  szFileReal = ResolveImageName(lpszFile);

  if ( (hr = LoadPictureFile(szFileReal, &ppic, &width, &height)) == ERROR_SUCCESS ) {
  
    free(szFileReal);
    bmih.biSize     = sizeof(BITMAPINFOHEADER);
    bmih.biWidth    = width;
    bmih.biHeight   = height;
    bmih.biPlanes   = 1;
    bmih.biBitCount = 32;
    bmih.biCompression = BI_RGB;
    bmih.biSizeImage = 0;
    bmih.biXPelsPerMeter = 0;
    bmih.biYPelsPerMeter = 0;
    bmih.biClrUsed = 0;
    bmih.biClrImportant = 0;

    ZeroMemory(&bmi,sizeof(BITMAPINFO));
    bmi.bmiHeader = bmih;
     
    ppv = (RGBQUAD*)malloc(sizeof(RGBQUAD)*width*height);

    if (pWidth) {
      *pWidth = width;
    }

    if (pHeight) {
      *pHeight = height;
    }

    hr = ppic->get_Type(&picType);

    if ( ppv != NULL && SUCCEEDED(hr) && picType == PICTYPE_BITMAP ) {
      hr = ppic->get_Handle((OLE_HANDLE*)&hBitmap);
      int sLines = GetDIBits(hdc, hBitmap, 0, height, ppv, &bmi, 0);

      ppic->Release();
      return (int)ppv;
    } else {
      return 0;
    }
  } else {
    free(szFileReal);
    return 0;
  }
}

// This function loads a file into an IStream.
HRESULT
LoadPictureFile
     ( LPCTSTR szFile
     , IPicture** ppIPicture
     , int* pWidth
     , int* pHeight
     )
{
  HANDLE hFile;
  DWORD dwFileSize;
  LPVOID pvData = NULL;
  HDC hdc = GetDC(NULL);
  
  *ppIPicture = NULL;

  // open file
  hFile = CreateFile(szFile, GENERIC_READ, 0, NULL, OPEN_EXISTING, 0, NULL);

  if ( hFile == INVALID_HANDLE_VALUE) {
    return HRESULT_FROM_WIN32(ERROR_FILE_NOT_FOUND);
  }
  // get file size
  dwFileSize = GetFileSize(hFile, NULL);
  if ( dwFileSize == -1) {
    return HRESULT_FROM_WIN32(GetLastError());
  }

  // alloc memory based on file size
  HGLOBAL hGlobal = GlobalAlloc(GMEM_MOVEABLE, dwFileSize);
  
  if ( hGlobal == NULL ) {
    return HRESULT_FROM_WIN32(ERROR_OUTOFMEMORY);
  }

  pvData = GlobalLock(hGlobal);
  if ( pvData == NULL ) {
    GlobalFree(hGlobal);
    return HRESULT_FROM_WIN32(ERROR_OUTOFMEMORY);
  }

  DWORD dwBytesRead = 0;
  // read file and store in global memory
  BOOL bRead = ReadFile(hFile, pvData, dwFileSize, &dwBytesRead, NULL);

  if ( bRead == FALSE ) {
    GlobalFree(hGlobal);
    // 'right' HRESULT ?
    return HRESULT_FROM_WIN32(ERROR_CANT_ACCESS_FILE);
  }
  GlobalUnlock(hGlobal);
  CloseHandle(hFile);

  LPSTREAM pstm = NULL;
  // create IStream* from global memory
  HRESULT hr = CreateStreamOnHGlobal(hGlobal, TRUE, &pstm);
  
  if ( !SUCCEEDED(hr) || !pstm ) {
    return hr;
  }

  // Create IPicture from image file
  hr = ::OleLoadPicture(pstm, dwFileSize, TRUE, IID_IPicture, (LPVOID *)ppIPicture);
  pstm->Release();
  if ( !SUCCEEDED(hr) ) {
    return hr;
  } else {
     // get width and height of picture
    long hmWidth;
    long hmHeight;
    (*ppIPicture)->get_Width(&hmWidth);
    (*ppIPicture)->get_Height(&hmHeight);
    // convert himetric to pixels
    *pWidth  = MulDiv(hmWidth, GetDeviceCaps(hdc, LOGPIXELSX), HIMETRIC_INCH);
    *pHeight = MulDiv(hmHeight, GetDeviceCaps(hdc, LOGPIXELSY), HIMETRIC_INCH);

    return S_OK;
  }
}

}
