%
% (c) The Foo Project, University of Glasgow, 1998
%

Stubs for marshalling and unmarshalling primitive
types.

\begin{code}
module HDirect 
	(
	  module HDirect

{- BEGIN_GHC_ONLY
	, Int64
   END_GHC_ONLY -}
	, Int32
	, Int16
	, Int8

	, Word8
	, Word16
	, Word32
	, Word64

	, Addr
	, Char
	, Wchar_t
	, Double
	, Float
	, Bool
{- BEGIN_GHC_ONLY
	, Hyper
   END_GHC_ONLY -}
	
	, intToInt32
	, intToWord32
	, word16ToWord32
	, intToWord16
	, intToInt16
	, toInt32
	
	, sizeofWord32
	, sizeofWord16
	, sizeofWord8
	, sizeofWord64

	, sizeofInt32
	, sizeofInt16
	, sizeofInt8
	, sizeofInt64
	
	, Ptr
	, castPtr
	
	) where

{- BEGIN_NOT_FOR_GHC -}
import Prelude hiding (readFloat)
{- END_NOT_FOR_GHC -}
import Char
import Int  
{- BEGIN_GHC_ONLY
        hiding ( sizeofInt64 )
   END_GHC_ONLY -}
import Addr
import Word
import Monad
import Pointer
import IOExts ( unsafePerformIO )
import Foreign   
	( ForeignObj
        , makeForeignObj
	, StablePtr
{- BEGIN_GHC_ONLY
	, foreignObjToAddr
	, mkForeignObj
   END_GHC_ONLY -}
	)
{- BEGIN_NOT_FOR_GHC -}
import -- don't bother gen. a dep. on this one, pal.
       AddrBits
{- END_NOT_FOR_GHC -}

{- BEGIN_GHC_ONLY
-- the next two are needed to implement enumToFlag, flagToIntTag.
import Bits
import GlaExts
   END_GHC_ONLY -}

\end{code}

At the moment the IDL compiler will emit calls to types with identity marshallers
(i.e., by-value marshallers for primitive, FFI-recognised types), so we need
to provide stubs for these here.


Int* marshalling functions:

\begin{code}
marshallInt :: Int -> IO Int
marshallInt x = return x

unmarshallInt :: Int -> IO Int
unmarshallInt x = return x

{- BEGIN_GHC_ONLY
writeInt :: Ptr Int -> Int -> IO ()
writeInt ptr v = writeIntOffAddr ptr 0 v

readInt :: Ptr Int -> IO Int
readInt ptr = readIntOffAddr ptr 0
   END_GHC_ONLY -}

--ToDo: generate host-specific versions of 
sizeofInt :: Word32
sizeofInt = 4

--Int8
marshallInt8 :: Int8 -> IO Int8
marshallInt8 v = return v

unmarshallInt8 :: Int8 -> IO Int8
unmarshallInt8 v = return v

writeInt8 :: Ptr Int8 -> Int8 -> IO ()
writeInt8 ptr v = writeInt8OffAddr ptr 0 v

readInt8 :: Ptr Int8 -> IO Int8
readInt8 ptr = readInt8OffAddr ptr 0

{- BEGIN_NOT_FOR_GHC -}
sizeofInt8 :: Word32
sizeofInt8 = 1
{- END_NOT_FOR_GHC -}

--Int16
marshallInt16 :: Int16 -> IO Int16
marshallInt16 x = return x
unmarshallInt16 :: Int16 -> IO Int16
unmarshallInt16 x = return x

writeInt16 :: Ptr Int16 -> Int16 -> IO ()
writeInt16 ptr v = writeInt16OffAddr ptr 0 v

readInt16 :: Ptr Int16 -> IO Int16
readInt16 ptr = readInt16OffAddr ptr 0

{- BEGIN_NOT_FOR_GHC -}
sizeofInt16 :: Word32
sizeofInt16 = 2
{- END_NOT_FOR_GHC -}

-- Int32
marshallInt32 :: Int32 -> IO Int32
marshallInt32 x = return x

unmarshallInt32 :: Int32 -> IO Int32
unmarshallInt32 x = return x

writeInt32 :: Ptr Int32 -> Int32 -> IO ()
writeInt32 ptr v = writeInt32OffAddr ptr 0 v

readInt32 :: Ptr Int32 -> IO Int32
readInt32 ptr = readInt32OffAddr ptr 0

{- BEGIN_NOT_FOR_GHC -}
sizeofInt32 :: Word32
sizeofInt32 = 4
{- END_NOT_FOR_GHC -}

{- BEGIN_GHC_ONLY
marshallInt64 :: Int64 -> IO Int64
marshallInt64 x = return x

unmarshallInt64 :: Int64 -> IO Int64
unmarshallInt64 x = return x

writeInt64 :: Ptr Int64 -> Int64 -> IO ()
writeInt64 ptr v = writeInt64OffAddr ptr 0 v

readInt64 :: Ptr Int64 -> IO Int64
readInt64 ptr = readInt64OffAddr ptr 0
 END_GHC_ONLY -}

sizeofInt64 :: Word32
sizeofInt64 = 8

{- BEGIN_GHC_ONLY
type Hyper   = Int64
marshallHyper   = marshallInt64
unmarshallHyper = unmarshallInt64
writeHyper      = writeInt64
readHyper       = readInt64
sizeofHyper     = sizeofInt64
   END_GHC_ONLY -}

writeInteger :: Ptr Integer -> Integer -> IO ()
writeInteger ptr x = 
{- BEGIN_GHC_ONLY
   writeInt64 ptr (integerToInt64 x)
   END_GHC_ONLY -}
{- BEGIN_NOT_FOR_GHC -}
   writeI64 ptr (toInt lo) (toInt hi)
  where
   (hi,lo) = x `divMod` (toInteger (maxBound :: Int) + 1)
{- END_NOT_FOR_GHC -}

readInteger :: Ptr Integer -> IO Integer
readInteger ptr = do
{- BEGIN_GHC_ONLY
  x <- readInt64 ptr
  return (int64ToInteger x)
   END_GHC_ONLY -}
{- BEGIN_NOT_FOR_GHC -}
  px <- readI64 ptr
  lo <- readInt32OffAddr px 0
  hi <- readInt32OffAddr px 1
  free ptr
  return ((toInteger lo) + (toInteger hi) * (toInteger (maxBound :: Int) + 1))
{- END_NOT_FOR_GHC -}

marshallInteger :: Integer -> IO (Int32, Int32)
marshallInteger i =  return (fromInteger lo, fromInteger hi)
 where
   (hi,lo) = i `divMod` (toInteger (maxBound :: Int) + 1)

unmarshallInteger :: (Int32,Int32) -> IO Integer
unmarshallInteger (hi,lo) =  return ((toInteger lo) + (toInteger hi) * (toInteger (maxBound :: Int) + 1))

marshallUInteger = marshallInteger
unmarshallUInteger = unmarshallInteger
readUInteger = readInteger
writeUInteger = writeInteger

\end{code}

Characters and bytes:

\begin{code}
marshallChar :: Char -> IO Char
marshallChar x = return x

unmarshallChar :: Char -> IO Char
unmarshallChar x = return x

writeChar :: Ptr Char -> Char -> IO ()
writeChar ptr v = writeCharOffAddr ptr 0 v

readChar :: Ptr Char -> IO Char
readChar ptr = readCharOffAddr ptr 0

sizeofChar :: Word32
sizeofChar     = 1

-- wide chars.
type Wchar_t = Word16

marshallWchar_t   = marshallWord16
unmarshallWchar_t = unmarshallWord16
writeWchar_t      = writeWord16
readWchar_t       = readWord16
sizeofWchar_t     = sizeofWord16


type Octet   = Byte
type Byte    = Word8

marshallByte   = marshallWord8
unmarshallByte = unmarshallWord8
writeByte      = writeWord8
readByte       = readWord8
sizeofByte     = sizeofWord8
\end{code}

Words:

\begin{code}
-- Word8:
marshallWord8 :: Word8 -> IO Word8
marshallWord8 x = return x

unmarshallWord8 :: Word8 -> IO Word8
unmarshallWord8 x = return x

writeWord8 :: Ptr Word8 -> Word8 -> IO ()
writeWord8 ptr v = writeWord8OffAddr ptr 0 v

readWord8 :: Ptr Word8 -> IO Word8
readWord8 ptr = readWord8OffAddr ptr 0

{- BEGIN_NOT_FOR_GHC -}
sizeofWord8 :: Word32
sizeofWord8 = 2
{- END_NOT_FOR_GHC -}

-- Word16:
marshallWord16 :: Word16 -> IO Word16
marshallWord16 x = return x

unmarshallWord16 :: Word16 -> IO Word16
unmarshallWord16 x = return x

writeWord16 :: Ptr Word16 -> Word16 -> IO ()
writeWord16 ptr v = writeWord16OffAddr ptr 0 v

readWord16 :: Ptr Word16 -> IO Word16
readWord16 ptr = readWord16OffAddr ptr 0

{- BEGIN_NOT_FOR_GHC -}
sizeofWord16 :: Word32
sizeofWord16 = 2
{- END_NOT_FOR_GHC -}

-- Word32:
marshallWord32 :: Word32 -> IO Word32
marshallWord32 x = return x

unmarshallWord32 :: Word32 -> IO Word32
unmarshallWord32 x = return x

writeWord32 :: Ptr Word32 -> Word32 -> IO ()
writeWord32 ptr v = writeWord32OffAddr ptr 0 v

readWord32 :: Ptr Word32 -> IO Word32
readWord32 ptr = readWord32OffAddr ptr 0

{- BEGIN_NOT_FOR_GHC -}
-- get this from Word now..
sizeofWord32 :: Word32
sizeofWord32 = 4
{- END_NOT_FOR_GHC -}

marshallWord64 :: Word64 -> IO Word64
marshallWord64 x = return x

unmarshallWord64 :: Word64 -> IO Word64
unmarshallWord64 x = return x

writeWord64 :: Ptr Word64 -> Word64 -> IO ()
readWord64 :: Ptr Word64 -> IO Word64
{- BEGIN_NOT_FOR_GHC -}
writeWord64 _ _ = undefined
readWord64 _ = undefined
{- END_NOT_FOR_GHC -}
{- BEGIN_GHC_ONLY
writeWord64 p v = writeWord64OffAddr p 0 v
readWord64 p = readWord64OffAddr p 0
   END_GHC_ONLY -}

{- BEGIN_NOT_FOR_GHC -}
sizeofWord64 :: Word32
sizeofWord64 = 8
{- END_NOT_FOR_GHC -}

\end{code}

Addr marshallers:

\begin{code}
marshallAddr :: Ptr Addr -> IO Addr
marshallAddr p = return p

unmarshallAddr :: Ptr Addr -> IO Addr
unmarshallAddr p = return p

writeAddr :: Ptr Addr -> Addr -> IO ()
{- BEGIN_DEBUG
writeAddr ptr a | ptr == nullAddr = ioError (userError "writeAddr: NULL pointer")
   END_DEBUG -}
writeAddr ptr a = writeAddrOffAddr ptr 0 a

readAddr :: Ptr Addr -> IO Addr
readAddr a = readAddrOffAddr a 0

sizeofAddr :: Word32
sizeofAddr     = 4

freeAddr = free
\end{code}

Double marshallers:

\begin{code}
marshallDouble :: Double -> IO Double
marshallDouble x = return x

unmarshallDouble :: Double -> IO Double
unmarshallDouble x = return x

writeDouble :: Ptr Double -> Double -> IO ()
writeDouble ptr x = writeDoubleOffAddr ptr 0 x

readDouble :: Ptr Double -> IO Double
readDouble ptr = readDoubleOffAddr ptr 0

sizeofDouble :: Word32
sizeofDouble   = 8

writeFloat :: Ptr Float -> Float -> IO ()
writeFloat ptr x = writeFloatOffAddr ptr 0 x

readFloat :: Ptr Float -> IO Float
readFloat ptr = readFloatOffAddr ptr 0

sizeofFloat :: Word32
sizeofFloat   = 4

\end{code}

Booleans - represented externally by a long (Int32):

\begin{code}
marshallBool :: Bool -> IO Int32
marshallBool v = marshallInt32 (if v then (-1) else 0)

unmarshallBool :: Int32 -> IO Bool
unmarshallBool v = return (v /= 0)

writeBool :: Ptr Bool -> Bool -> IO ()
writeBool ptr v = writeInt32 (castPtr ptr) (if v then (-1) else 0)

readBool :: Ptr Int32 -> IO Bool
readBool ptr = do
  v <- readInt32 ptr
  return ( v /= 0 )

sizeofBool :: Word32
sizeofBool = sizeofInt32
\end{code}

\begin{code}
addNCastPtr :: Ptr a -> Word32 -> Ptr b
addNCastPtr v off = v `plusAddr` (toInt off)

mkPtr :: Addr -> Ptr a
mkPtr a = a

pointerToAddr :: Ptr a -> Addr
pointerToAddr p = p

derefPtr :: Ptr a -> IO (Ptr b)
derefPtr p = readAddrOffAddr p 0

indexPtr :: Ptr Addr -> Int -> IO Addr
indexPtr p off = readAddrOffAddr p off
\end{code}

The unit of allocation is an 8-bit byte.

\begin{code}
allocOutPtr :: IO (Ptr a)
allocOutPtr = alloc 4 -- 4 bytes in a pointer. ToDo: 64-bit platform'ify.

allocBytes :: Int -> IO (Ptr a)
allocBytes len = alloc (intToWord32 len)

allocWords :: Int -> IO (Ptr a)
allocWords len = alloc (4 * intToWord32 len)

{- BEGIN_GHC_ONLY
foreign import "malloc" alloc :: Word32 -> IO (Ptr a)
foreign import "free"   free  :: Ptr a  -> IO ()
   END_GHC_ONLY -}

{- BEGIN_NOT_FOR_GHC -}
alloc :: Word32 -> IO (Ptr a)
free :: Ptr a -> IO ()
alloc len = primMalloc len
free ptr = primFree ptr
{- END_NOT_FOR_GHC -}

doThenFree ::(Ptr a -> IO ()) -> (Ptr a -> IO b) -> Ptr a -> IO b
doThenFree f act ptr = do
   v <- act ptr
   f ptr
   return v

trivialFree :: a -> IO ()
trivialFree _ = return ()

fillIn :: Int -> (Ptr a -> IO ()) -> IO (Ptr a)
fillIn sz refreemarshall = do
  ptr <- allocBytes sz
  refreemarshall ptr
  return ptr

\end{code}

[ptr]Ptr marshalling

\begin{code}
marshallptr :: Ptr a -> IO (Ptr a)
marshallptr v = return v

unmarshallptr :: Ptr a -> IO (Ptr a)
unmarshallptr v = return v

writeptr :: Ptr Addr -> Addr -> IO ()
writeptr a v = writeAddrOffAddr a 0 v

readptr :: Ptr a -> IO (Ptr b)
readptr a = readAddrOffAddr a 0

writefptr :: Ptr ForeignObj -> ForeignObj -> IO ()
writefptr p v = writeForeignObjOffAddr p 0 v
\end{code}

[unique]Ptr marshalling

\begin{code}
marshallunique :: (IO (Ptr a))
               -> (Ptr a -> a -> IO ())
	       -> Maybe a
	       -> IO (Ptr a)
marshallunique alloc marshallInto mb = 
  case mb of
    Nothing -> return nullPtr
    Just x  -> marshallref alloc marshallInto x

marshallMaybe :: (a -> IO b) -> b -> Maybe a -> IO b
marshallMaybe mshall def Nothing  = return def
marshallMaybe mshall def (Just x) = mshall x

writeMaybe :: (Ptr a -> a -> IO ())
           -> Ptr (Maybe a)
	   -> Maybe a
	   -> IO ()
writeMaybe wr ptr Nothing  = writeptr ptr nullPtr
writeMaybe wr ptr (Just x) = wr ptr x

readMaybe :: (Ptr a -> IO b) -> Ptr a -> IO (Maybe b)
readMaybe rd ptr
 | ptr == nullAddr = return Nothing
 | otherwise       = rd ptr >>= return . Just

writeunique :: IO (Ptr a)
	    -> (Ptr a -> a -> IO ())
	    -> Ptr (Maybe a)
	    -> Maybe a
	    -> IO ()
writeunique alloc marshallInto p mb =
  case mb of
    Nothing  -> writeptr p nullPtr
    Just x   -> writeref alloc marshallInto p x

unmarshallunique :: (Ptr a -> IO a) -> Ptr a -> IO (Maybe a)
unmarshallunique refUnmarshall ptr
 | ptr == nullPtr = return Nothing
 | otherwise      = do
    x <- unmarshallref refUnmarshall ptr
    return (Just x)


readunique :: (Ptr a -> IO a) -> Ptr (Ptr a) -> IO (Maybe a)
readunique refUnmarshall ptr
 | ptr == nullPtr = return Nothing
 | otherwise      = do
   v <- readptr ptr
   if nullPtr == v then
      return Nothing
    else do
      x <- refUnmarshall v
      return (Just x)

freeunique :: (Ptr a -> IO ()) -> Ptr (Ptr a) -> IO ()
freeunique freeptee ptr
 | ptr == nullAddr = return ()
 | otherwise       = do
    ptr' <- derefPtr ptr
    freeptee ptr'
    free ptr

\end{code}

Marshalling [unique]void* pointers

\begin{code}
marshallunique_ptr :: Maybe Addr -> IO (Ptr Addr)
marshallunique_ptr mb = 
  case mb of
    Nothing -> marshallptr nullPtr
    Just x  -> marshallptr x
\end{code}


[ref]Ptr marshalling

\begin{code}
marshallref :: (IO (Ptr a)) -> (Ptr a -> a -> IO ()) -> a -> IO (Ptr a)
marshallref alloc marshallInto v = do
   px <- alloc
   marshallInto px v
   return px

writeref :: (IO (Ptr a)) -> (Ptr a -> a -> IO ()) -> Ptr Addr -> a -> IO ()
writeref alloc marshallInto ptr v = do
  px <- marshallref alloc marshallInto v
  writeptr ptr px

unmarshallref :: (Ptr a -> IO b) -> Ptr a -> IO b
unmarshallref refUnmarshall ptr = refUnmarshall ptr

readref :: (Ptr a -> IO a) -> Ptr Addr -> IO a
readref refUnmarshall ptr = do
  px <- readptr ptr
  unmarshallref refUnmarshall (castPtr px)

freeref :: (Ptr a -> IO ()) -> Ptr b -> IO ()
freeref free_inner pptr = do
   ptr <- readAddr pptr
   free_inner ptr
   free pptr

\end{code}


All by-reference marshalling and unmarshalling functions
of enums can be expressed using these two stubs:

\begin{code}
writeenum16 :: (b -> IO Int16) -> Ptr Int16 -> b -> IO ()
writeenum16 marshall pv v =
  marshall v >>= \ x ->
  writeInt16 pv x

readenum16 :: (Int16 -> IO a) -> Ptr (Int16) -> IO a
readenum16 unmarshall pv = do
  v <- readInt16 pv
  unmarshall v

marshallEnum16 :: Enum a => a -> IO Int16
marshallEnum16 v = return (intToInt16 (fromEnum v))

unmarshallEnum16 :: Enum a => Int16 -> IO a
unmarshallEnum16 x = return (toEnum (int16ToInt x))

marshallEnum32 :: Enum a => a -> IO Int32
marshallEnum32 v = return (intToInt32 (fromEnum v))

unmarshallEnum32 :: Enum a => Int32 -> IO a
unmarshallEnum32 x = return (toEnum (int32ToInt x))

writeEnum32 :: Enum a => Ptr Int32 -> a -> IO ()
writeEnum32 p v = writeInt32 p (intToInt32 (fromEnum v))

readEnum32 :: Enum a => Ptr Int32 -> IO a
readEnum32 p = do
  x <- readInt32 p 
  return (toEnum (int32ToInt x))

writeEnum16 :: Enum a => Ptr Int16 -> a -> IO ()
writeEnum16 p v = writeInt16 p (intToInt16 (fromEnum v))

readEnum16 :: Enum a => Ptr Int16 -> IO a
readEnum16 p = do
  x <- readInt16 p 
  return (toEnum (int16ToInt x))

\end{code}

\begin{code}
marshalllist :: Word32
	     -> (Ptr a -> a -> IO ())
	     -> [a]
	     -> IO (Ptr [a])
marshalllist szof writeelt ls = do
 arr <- alloc (len*szof)
 foldM writeElt (castPtr arr) ls
 return arr
  where
   len = intToWord32 (length ls)

   writeElt addr v = do
     writeelt addr v
     return (addNCastPtr addr szof)

unmarshalllist :: Word32 -> Word32 -> Word32 -> (Ptr a -> IO a) -> Ptr [a] -> IO [a]
unmarshalllist szof offset len unpack ptr = do
 let ptr0 = addNCastPtr ptr (offset*szof)
 loop ptr0 len
  where
   loop ptr 0 = return []
   loop ptr n = do
    v  <- unpack ptr
    let ptr' = addNCastPtr ptr szof
    vs <- loop ptr' (n-1)
    return (v:vs)

unmarshallSingle :: (Ptr a -> IO a) -> Ptr a -> IO [a]
unmarshallSingle ref ptr 
 | ptr == nullPtr = return []
 | otherwise      = do
      x <- ref ptr
      return [x]

writelist :: Bool -> Word32 -> (Ptr a -> a -> IO ()) -> Ptr [a] -> [a] -> IO ()
writelist do_alloc szof writeelt pptr ls = do
 the_ptr <- 
    (if do_alloc then do
        ptr <- alloc (szof * fromInt len)
	writeAddr pptr ptr
	return ptr
      else
        return (castPtr pptr))
 foldM writeElt the_ptr ls
 return ()
  where
   len = length ls

   writeElt addr v = do
    writeelt addr v
    return (addNCastPtr addr szof)

readlist :: Word32 -> Word32 -> (Ptr a -> IO a) -> Ptr [a] -> IO [a]
readlist szof len unpack ptr = do
 let ptr0 = castPtr ptr
 loop ptr0 len
  where
   loop ptr 0 = return []
   loop ptr n = do
    v  <- unpack ptr
    let ptr' = addNCastPtr ptr szof
    vs <- loop ptr' (n-1)
    return (v:vs)

freelist :: Word32 -> Word32 -> (Ptr a -> IO ()) -> Ptr [a] -> IO ()
freelist szof len free_elt ptr = do
	go (castPtr ptr) len
	free ptr
  where
    go pptr 0 = return ()
    go pptr x = do
       ptr <- readAddr pptr
       free_elt ptr
       let pptr' = addNCastPtr pptr szof
       go pptr' (x-1)

\end{code}

Unpacking null terminated character strings:

\begin{code}
marshallString :: String -> IO (Ptr String)
marshallString str = do
 pstr  <- alloc (len*sizeofChar)
 pstr' <- foldM writeElt pstr str
 writeChar pstr' '\0'
 return pstr
  where
   len = intToWord32 (length str + 1)

   writeElt addr v = do
     writeChar addr v
     return (addNCastPtr addr sizeofChar)

marshallBString :: Int -> String -> IO (Ptr String)
marshallBString len str = do
 pstr  <- alloc (len'*sizeofChar)
 pstr' <- foldM writeElt pstr (take len str)
 writeChar pstr' '\0'
 return pstr
  where
   len' = intToWord32 (len + 1)

   writeElt addr v = do
     writeChar addr v
     return (addNCastPtr addr sizeofChar)

unmarshallString :: Ptr String -> IO String
unmarshallString ptr
 | ptr == nullAddr = return ""
 | otherwise	   = do
   let ptr0 = addNCastPtr ptr 0
   loop ptr0
  where
   loop ptr = do
    v  <- readChar ptr
    if v == '\0'
     then return []
     else do
       let ptr' = addNCastPtr ptr sizeofChar
       vs <- loop ptr'
       return (v:vs)

-- at most len chars. or zero terminated.
unmarshallBString :: Int -> Ptr String -> IO String
unmarshallBString len ptr
 | ptr == nullAddr = return ""
 | otherwise	   = do
   let ptr0 = addNCastPtr ptr 0
   loop ptr0 0
  where
   loop ptr n | n > len = return ""
   loop ptr n = do
    v  <- readChar ptr
    if v == '\0'
     then return []
     else do
       let ptr' = addNCastPtr ptr sizeofChar
       vs <- loop ptr' (n+1)
       return (v:vs)

readString :: Ptr (Ptr String) -> IO String
readString pstr = do
  ptr <- readAddr pstr
  unmarshallString ptr

readBString :: Int -> Ptr (Ptr String) -> IO String
readBString len pstr = do
  ptr <- readAddr pstr
  unmarshallBString len ptr

writeString :: Bool -> Ptr String -> String -> IO ()
writeString do_alloc ppstr str = do
  pstr <-
    if not do_alloc then
       return (castPtr ppstr)
     else do
       arr <- alloc (fromInt string_len)
       writeAddr ppstr arr
       return arr
  pstr' <- foldM writeElt pstr str
  writeChar pstr' '\0'
 where
   string_len = length str + 1 {- terminator -}

   writeElt addr v = do
     writeChar addr v
     return (addNCastPtr addr sizeofChar)

writeBString :: Bool -> Int -> Ptr a -> String -> IO ()
writeBString do_alloc len ptr str = do
  pstr  <-
    if do_alloc then
        alloc (fromInt len * sizeofChar)
    else
        return (castPtr ptr)
  pstr' <- foldM writeElt pstr (take len str)
  writeChar pstr' '\0'
 where
   writeElt addr v = do
     writeChar addr v
     return (addNCastPtr addr sizeofChar)

freeString = free
\end{code}

\begin{code}
-- at most len elements
marshallblist :: Word32 -> Word32 -> (Ptr a -> a -> IO ()) -> [a] -> IO (Ptr [a])
marshallblist szof len writeelt ls = do
 arr <- alloc (len'*szof)
 foldM writeElt (castPtr arr) ls
 return arr
  where
   len' = atLeast len (intToWord32 0) ls

   atLeast len n _ | len == n = len
   atLeast len n [] = n
   atLeast len n (x:xs) = atLeast len (n+1) xs

   writeElt addr v = do
    writeelt addr v
    return (addNCastPtr addr szof)

writeblist :: Word32 -> Word32 -> (Ptr a -> a -> IO ()) -> Ptr [a] -> [a] -> IO ()
writeblist szof len writeelt ptr ls = do
 foldM writeElt ptr (take (toInt len) ls)
 return ()
  where

   writeElt addr v = do
    writeelt addr v
    return (addNCastPtr addr szof)

readblist :: Word32 -> Word32 -> (Ptr a -> IO a) -> Ptr a -> IO [a]
readblist szof len unpack ptr = do
 let ptr0 = castPtr ptr
 loop ptr0 len
  where
   loop ptr 0 = return []
   loop ptr n = do
    v  <- unpack ptr
    let ptr' = addNCastPtr ptr szof
    vs <- loop ptr' (n-1)
    return (v:vs)

\end{code}

Misc coercion functions/shortcuts:
(ToDo: try to avoid generating them in the first place!)

\begin{code}
word16ToInt32 :: Word16 -> Int32
word16ToInt32 w = intToInt32 (word16ToInt w)

--HACK - not correct!!!!
word32ToInt32 :: Word32 -> Int32
word32ToInt32 w = intToInt32 (word32ToInt w)

int32ToWord32 :: Int32 -> Word32
int32ToWord32 w = intToWord32 (int32ToInt w)

int16ToWord32 :: Int16 -> Word32
int16ToWord32 w = intToWord32 (int16ToInt w)

intToChar :: Int -> Char
intToChar = chr

charToInt32 :: Char -> Int32
charToInt32 c = intToInt32 (ord c)

word32ToChar :: Word32 -> Char
word32ToChar w = chr (toInt w)

charToWord32 :: Char -> Word32
charToWord32 c = intToWord32 (ord c)

\end{code}

\begin{code}
toInt32 :: Integral a => a -> Int32
toInt32 i = intToInt32 (toInt i)

toInt16 :: Integral a => a -> Int16
toInt16 i = intToInt16 (toInt i)
\end{code}

ForeignObj marshallers:

\begin{code}
marshallForeignObj :: ForeignObj -> IO ForeignObj
marshallForeignObj x = return x

unmarshallForeignObj :: ForeignObj -> IO ForeignObj
unmarshallForeignObj x = return x

writeForeignObj :: Ptr ForeignObj -> ForeignObj -> IO ()
writeForeignObj ptr fo = do
   a <- foreignObjToAddr fo
   writeAddrOffAddr ptr 0 a

-- a C pointer, really.
sizeofForeignObj :: Word32
sizeofForeignObj = sizeofAddr

nullForeignObj :: ForeignObj
{- BEGIN_GHC_ONLY
nullForeignObj = unsafePerformIO (mkForeignObj nullAddr)
   END_GHC_ONLY -}
{- BEGIN_NOT_FOR_GHC -}
nullForeignObj = unsafePerformIO (makeForeignObj nullAddr finalNoFree)
{- END_NOT_FOR_GHC -}

nullFinaliser :: Addr
nullFinaliser = finalNoFree
\end{code}

\begin{code}
readStablePtr :: Ptr (StablePtr a) -> IO (StablePtr a)
readStablePtr ptr = readStablePtrOffAddr ptr 0
\end{code}

Default by-value 'marshallers' for structs and unions.

\begin{code}
marshallStruct :: String -> a -> IO b
marshallStruct nm _ = ioError (userError (nm ++ ": Marshalling structs by value is not supported yet."))

unmarshallStruct :: String -> a -> IO c
unmarshallStruct nm _ = ioError (userError (nm ++ ": Unmarshalling structs by value is not supported yet."))

marshallUnion :: String -> a -> IO b
marshallUnion nm _ = ioError (userError (nm ++ ": Marshalling unions by value is not supported yet."))

unmarshallUnion :: String -> a -> b -> IO c
unmarshallUnion nm _ _ = ioError (userError (nm ++ ": Unmarshalling unions by value is not supported yet."))
\end{code}

Ptr marshallers:

\begin{code}
marshallPointer :: Ptr a -> IO Addr
marshallPointer p = return p

unmarshallPointer :: Addr -> IO (Ptr a)
unmarshallPointer ptr = makeNoFreePtr ptr

writePointer :: Ptr (Ptr a) -> Ptr a -> IO ()
writePointer = writeAddr

readPointer :: Ptr (Ptr a) -> IO (Ptr a)
readPointer ptr = do
  x <- readAddr ptr
  makeNoFreePtr x

sizeofPtr :: Word32
sizeofPtr = sizeofAddr
\end{code}

\begin{code}
primInvokeIt :: (Addr -> Addr -> IO a) -> Int -> IO Addr -> IO a
primInvokeIt meth offset mk_obj_ptr = do
  obj_ptr <- mk_obj_ptr 
  lpVtbl  <- derefPtr obj_ptr
  methPtr <- indexPtr lpVtbl offset
  meth methPtr obj_ptr

primInvokeItFO :: (Addr -> ForeignObj -> IO a) -> Int -> IO ForeignObj -> IO a
primInvokeItFO meth offset mk_obj_ptr = do
  obj_ptr <- mk_obj_ptr 
  a       <- foreignObjToAddr obj_ptr
  lpVtbl  <- derefPtr a
  methPtr <- indexPtr lpVtbl offset
  meth methPtr obj_ptr
\end{code}

\begin{code}
stackStringLen :: Int -> String -> (Ptr String -> IO a) -> IO a
stackStringLen len str f
      = let slen = length str + 1 `max` len
        in stackFrame (fromInt slen) $ \pstr -> do 
	 writeString False pstr str
         f pstr

\end{code}

\begin{code}
{- BEGIN_GHC_ONLY

enumToFlag :: Enum a => a -> Int
enumToFlag tg = toInt ((1::Word32) `shiftL` enumToInt tg)

enumToInt :: Enum a => a -> Int
enumToInt tg = I# (getTag# tg)

flagToIntTag :: Int -> Int
flagToIntTag x | x < 0     = error "flagToIntTag: negative tag"
               | otherwise = go 0 x
   -- could've used Prelude.logBase, but don't need the precision it offers.
 where
    go n 0 = n
    go n 1 = n + 1
    go n x = go (n + 1) (x `div` 2)

unboxInt :: Int -> Int#
unboxInt (I# x#) = x#

toIntFlag :: Int -> Int -> Int
toIntFlag b v = toInt ((1::Word32) `shiftL` (v + flagToIntTag b))

   END_GHC_ONLY -}
\end{code}
