module PointE where

import Exp

---- Point

-- Points are pairs, not expressions.  They're "syntactic", thanks to the
-- instance for pairs.  They're generally taken to be in rectangular
-- coordinates (X,Y)
type PointE = (FloatE, FloatE)

-- Polar coordinates (rho,theta) are sometimes more convenient
type PolarPointE = (FloatE, FloatE)

-- Sometimes useful for displacement or sizes
type VectorE = (FloatE,FloatE)

-- (Square of) distance from point to origin.
distOsq, distO, angleO :: PointE -> FloatE
distOsq (x,y) = x^2 + y^2
distO         = sqrt . distOsq

angleO (x,y) = atan2 y x

fromPolar :: PolarPointE -> PointE
fromPolar (rho,theta) = (rho * cos theta, rho * sin theta)

toPolar :: PointE -> PolarPointE
toPolar p = (distO p, angleO p)

i2fPair :: (IntE,IntE) -> PointE
i2fPair (i,j) = (i2f i, i2f j)

f2iPair :: PointE -> (IntE,IntE)
f2iPair (x,y) = (f2i x, f2i y)

